@@ -3,10 +3,11 @@ package Parse::Syslog;
 use Carp;
 use Symbol;
 use Time::Local;
+use IO::File;
 use strict;
 use vars qw($VERSION);
 
-$VERSION = '1.01';
+$VERSION = '1.05';
 
 my %months_map = (
     'Jan' => 0, 'Feb' => 1, 'Mar' => 2,
@@ -19,23 +20,53 @@ my %months_map = (
     'oct' => 9, 'nov' =>10, 'dec' =>11,
 );
 
-# fast timelocal
-my $str2time_last_time;
-my $str2time_last_day;
-my $str2time_last_month;
-my $enable_year_decrement = 1; # year-increment algorithm: if in january, if december is seen, decrement
-                               # year
-# 0: sec, 1: min, 2: h, 3: day, 4: month, 5: year
-sub str2time($$$$$$$)
+sub is_dst_switch($$$)
 {
+    my ($self, $t, $time) = @_;
+
+    # calculate the time in one hour and see if the difference is 3600 seconds.
+    # if not, we are in a dst-switch hour
+    # note that right now we only support 1-hour dst offsets
+
+    # cache the result
+    if(defined $self->{is_dst_switch_last_hour} and
+        $self->{is_dst_switch_last_hour} == $t->[3]<<5+$t->[2]) {
+        return @{$self->{is_dst_switch_result}};
+    }
+
+    # calculate a number out of the day and hour to identify the hour
+    $self->{is_dst_switch_last_hour} = $t->[3]<<5+$t->[2];
+
+    # let's see the timestamp in one hour
+    # 0: sec, 1: min, 2: h, 3: day, 4: month, 5: year
+    my $time_plus_1h = timelocal($t->[0], $t->[1], $t->[2]+1, $t->[3], $t->[4], $t->[5]);
+
+    if($time_plus_1h - $time > 4000) {
+        return 3600, $time-$time%3600+3600;
+    }
+    else {
+        return 0, undef;
+    }
+}
+
+# fast timelocal, cache minute's timestamp
+# don't cache more than minute because of daylight saving time switch
+# 0: sec, 1: min, 2: h, 3: day, 4: month, 5: year
+sub str2time($$$$$$$$)
+{
+    my $self = shift @_;
     my $GMT = pop @_;
-    my $day_secs = $_[2]*3600+$_[1]*60+$_[0];
-    if(defined $str2time_last_time) {
-        if( $_[3] == $str2time_last_day and
-            $_[4] == $str2time_last_month )
-        {
-            return $str2time_last_time + $day_secs;
-        }
+
+    my $lastmin = $self->{str2time_lastmin};
+    if(defined $lastmin and
+        $lastmin->[0] == $_[1] and
+        $lastmin->[1] == $_[2] and
+        $lastmin->[2] == $_[3] and
+        $lastmin->[3] == $_[4] and
+        $lastmin->[4] == $_[5])
+    {
+        $self->{last_time} = $self->{str2time_lastmin_time} + $_[0];
+        return $self->{last_time} + ($self->{dst_comp}||0);
     }
 
     my $time;
@@ -46,11 +77,30 @@ sub str2time($$$$$$$)
         $time = timelocal(@_);
     }
 
-    $str2time_last_time = $time - $day_secs;
-    $str2time_last_day = $_[3];
-    $str2time_last_month = $_[4];
+    # compensate for DST-switch
+    # - if a timewarp is detected (1:00 -> 1:30 -> 1:00):
+    # - test if we are in a DST-switch-hour
+    # - compensate if yes
+    if(!$GMT and !defined $self->{dst_comp} and
+        defined $self->{last_time} and
+        $self->{last_time}-$time > 1200 and
+        $self->{last_time}-$time < 3600)
+    {
+        my ($off, $until) = $self->is_dst_switch(\@_, $time);
+        if($off) {
+            $self->{dst_comp} = $off;
+            $self->{dst_comp_until} = $until;
+        }
+    }
+    if(defined $self->{dst_comp_until} and $time > $self->{dst_comp_until}) {
+        delete $self->{dst_comp};
+        delete $self->{dst_comp_until};
+    }
 
-    return $time;
+    $self->{str2time_lastmin} = [ @_[1..5] ];
+    $self->{str2time_lastmin_time} = $time-$_[0];
+    $self->{last_time} = $time;
+    return $time+($self->{dst_comp}||0);
 }
 
 sub _use_locale($)
@@ -75,15 +125,22 @@ sub new($$;%)
     if(not defined $data{year}) {
         $data{year} = (localtime(time))[5]+1900;
     }
+    $data{type} = 'syslog' unless defined $data{type};
     $data{_repeat}=0;
 
-    if(ref $file eq 'File::Tail') {
-        $data{filetail} = 1;
+    if(UNIVERSAL::isa($file, 'IO::Handle')) {
         $data{file} = $file;
     }
+    elsif(UNIVERSAL::isa($file, 'File::Tail')) {
+        $data{file} = $file;
+        $data{filetail}=1;
+    }
+    elsif(! ref $file) {
+        $data{file} = new IO::File($file, "<");
+        defined $data{file} or croak "can't open $file: $!";
+    }
     else {
-        $data{file}=gensym;
-        open($data{file}, "<$file") or croak "can't open $file: $!";
+        croak "argument must be either a file-name or an IO::Handle object.";
     }
 
     if(defined $data{locale}) {
@@ -101,6 +158,27 @@ sub new($$;%)
     return bless \%data, $class;
 }
 
+sub _year_increment($$)
+{
+    my ($self, $mon) = @_;
+
+    # year change
+    if($mon==0) {
+        $self->{year}++ if defined $self->{_last_mon} and $self->{_last_mon} == 11;
+        $self->{enable_year_decrement} = 1;
+    }
+    elsif($mon == 11) {
+        if($self->{enable_year_decrement}) {
+            $self->{year}-- if defined $self->{_last_mon} and $self->{_last_mon} != 11;
+        }
+    }
+    else {
+        $self->{enable_year_decrement} = 0;
+    }
+
+    $self->{_last_mon} = $mon;
+}
+
 sub _next_line($)
 {
     my $self = shift;
@@ -109,11 +187,11 @@ sub _next_line($)
         return $f->read;
     }
     else {
-        return <$f>;
+        return $f->getline;
     }
 }
 
-sub next($)
+sub _next_syslog($)
 {
     my ($self) = @_;
 
@@ -122,43 +200,39 @@ sub next($)
         return $self->{_repeat_data};
     }
 
-    line: while(my $str = $self->_next_line()) {
+    my $file = $self->{file};
+    line: while(my $str = $self->_next_line) {
         # date, time and host 
         $str =~ /^
-            (\S{3})\s+(\d+)   # date  -- 1, 2
+            (\S{3})\s+(\d+)      # date  -- 1, 2
             \s
-            (\d+):(\d+):(\d+) # time  -- 3, 4, 5
+            (\d+):(\d+):(\d+)    # time  -- 3, 4, 5
+            (?:\s<\w+\.\w+>)?    # FreeBSD's verbose-mode
             \s
-            ([-\w\.]+)        # host  -- 6
+            ([-\w\.\@:]+)        # host  -- 6
             \s+
-            (.*)              # text  -- 7
+            (?:\[LOG_[A-Z]+\]\s+)?  # FreeBSD
+            (.*)                 # text  -- 7
             $/x or do
         {
-            carp "line not in syslog format: $str";
+            warn "WARNING: line not in syslog format: $str";
             next line;
         };
         
         my $mon = $months_map{$1};
         defined $mon or croak "unknown month $1\n";
 
-        # year change
-        if($mon==0) {
-            $self->{year}++ if defined $self->{_last_mon} and $self->{_last_mon} == 11;
-            $enable_year_decrement = 1;
-        }
-        elsif($mon == 11) {
-            if($enable_year_decrement) {
-                $self->{year}-- if defined $self->{_last_mon} and $self->{_last_mon} != 11;
-            }
-        }
-        else {
-            $enable_year_decrement = 0;
-        }
-
-        $self->{_last_mon} = $mon;
+        $self->_year_increment($mon);
 
         # convert to unix time
-        my $time = str2time($5,$4,$3,$2,$mon,$self->{year}-1900,$self->{GMT});
+        my $time = $self->str2time($5,$4,$3,$2,$mon,$self->{year}-1900,$self->{GMT});
+        if(not $self->{allow_future}) {
+            # accept maximum one day in the present future
+            if($time - time > 86400) {
+                warn "WARNING: ignoring future date in syslog line: $str";
+                next line;
+            }
+        }
 
         my ($host, $text) = ($6, $7);
 
@@ -167,7 +241,7 @@ sub next($)
             next line if defined $self->{repeat} and not $self->{repeat};
             next line if not defined $self->{_last_data}{$host};
             $1 > 0 or do {
-                carp "last message repeated 0 or less times??";
+                warn "WARNING: last message repeated 0 or less times??\n";
                 next line;
             };
             $self->{_repeat}=$1-1;
@@ -194,7 +268,7 @@ sub next($)
             (.*)            # text      -- 6
             $/x or do
         {
-            carp "line not in syslog format: $str";
+            warn "WARNING: line not in syslog format: $str";
             next line;
         };
 
@@ -225,6 +299,80 @@ sub next($)
     return undef;
 }
 
+sub _next_metalog($)
+{
+    my ($self) = @_;
+    my $file = $self->{file};
+    line: while(my $str = $self->_next_line) {
+	# date, time and host 
+	
+	$str =~ /^
+            (\S{3})\s+(\d+)   # date  -- 1, 2
+            \s
+            (\d+):(\d+):(\d+) # time  -- 3, 4, 5
+	                      # host is not logged
+            \s+
+            (.*)              # text  -- 6
+            $/x or do
+        {
+            warn "WARNING: line not in metalog format: $str";
+            next line;
+        };
+	
+        my $mon = $months_map{$1};
+        defined $mon or croak "unknown month $1\n";
+
+        $self->_year_increment($mon);
+
+        # convert to unix time
+        my $time = $self->str2time($5,$4,$3,$2,$mon,$self->{year}-1900,$self->{GMT});
+	
+	my $text = $6;
+
+        $text =~ /^
+            \[(.*?)\]        # program   -- 1
+           	             # no PID
+	    \s+
+            (.*)             # text      -- 2
+            $/x or do
+        {
+	    warn "WARNING: text line not in metalog format: $text ($str)";
+            next line;
+        };
+
+        if($self->{arrayref}) {
+            return [
+                $time,  # 0: timestamp 
+                'localhost',  # 1: host      
+                $1,     # 2: program   
+                undef,  # 3: (no) pid
+                $2,     # 4: text
+                ];
+        }
+        else {
+            return {
+                timestamp => $time,
+                host      => 'localhost',
+                program   => $1,
+                text      => $2,
+            };
+        }
+    }
+    return undef;
+}
+
+sub next($)
+{
+    my ($self) = @_;
+    if($self->{type} eq 'syslog') {
+        return $self->_next_syslog();
+    }
+    elsif($self->{type} eq 'metalog') {
+        return $self->_next_metalog();
+    }
+    croak "Internal error: unknown type: $self->{type}";
+}
+
 1;
 
 __END__
@@ -235,10 +383,7 @@ Parse::Syslog - Parse Unix syslog files
 
 =head1 SYNOPSIS
 
- my $parser = Parse::Syslog->new( '/var/log/syslog'
-                                , year   => 2001
-                                , locale => qw(de_CH ru_RU.koi8r)
-                                );
+ my $parser = Parse::Syslog->new( '/var/log/syslog', year => 2001);
  while(my $sl = $parser->next) {
      ... access $sl->{timestamp|host|program|pid|text} ...
  }
@@ -255,15 +400,47 @@ with Unix-timestamp, host, program, pid and text returned in a hash-reference.
 
 =head2 Constructing a Parser
 
-B<new> requires as first argument a file-name for the syslog-file to be parsed.
-Alternatively, you can pass a File::Tail object as first argument, in which
+B<new> requires as first argument a source from where to get the syslog lines. It can
+be:
+
+=over 4
+
+=item *
+
+a file-name for the syslog-file to be parsed.
+
+=item *
+
+a File::Tail object as first argument, in which
 case the I<read> method will be called to get lines to process.
 
+=item *
+
+a file handle (GLOB-ref) for an already-opened syslog-file.
+
+=back
+
 After the file-name (or File::Tail object), you can specify options as a hash.
 The following options are defined:
 
 =over 8
 
+=item B<type>
+
+Format of the "syslog" file. Can be one of:
+
+=over 8
+
+=item I<syslog>
+
+Traditional "syslog" (default)
+
+=item I<metalog>
+
+Metalog (see http://metalog.sourceforge.net/)
+
+=back
+
 =item B<year>
 
 Syslog files usually do store the time of the event without year. With this
@@ -315,6 +492,12 @@ text
 Optional. Specifies an additional locale name or the array of locale names for
 the parsing of log files with national characters.
 
+=item B<allow_future>
+
+If true will allow for timestamps in the future. Otherwise timestamps of one day
+in the future and more will not be returned (as a safety measure against wrong
+configurations, bogus --year arguments, etc.)
+
 =back
 
 =head2 Parsing the file
@@ -395,6 +578,12 @@ David Schweikert <dws@ee.ethz.ch>
  2001-08-20 ds 0.03 implemented GMT option, year specification, File::Tail
  2001-10-31 ds 0.04 faster time parsing, implemented 'arrayref' option, better time-increment algorithm
  2002-01-29 ds 0.05 ignore -- MARK -- lines, low-case months, space in program names
+ 2002-05-02 ds 1.00 HP-UX fixes, parse 'above message repeats xx times'
+ 2002-05-25 ds 1.01 added support for localized month names (uchum@mail.ru)
+ 2002-10-28 ds 1.02 fix off-by-one-hour error when running during daylight saving time switch
+ 2004-01-19 ds 1.03 do not allow future dates (if allow_future is not true)
+ 2004-07-11 ds 1.04 added support for type 'metalog'
+ 2005-12-24 ds 1.05 allow passing of a IO::Handle object to new
 
 =cut
 
